"""Onzr: deezer client."""

import functools
import hashlib
import logging
from dataclasses import asdict, dataclass
from enum import IntEnum, StrEnum
from threading import Thread
from time import sleep
from typing import Generator, List, Protocol

import deezer
import requests
from Cryptodome.Cipher import Blowfish

from .config import settings

logger = logging.getLogger(__name__)


class StreamQuality(StrEnum):
    """Track stream quality."""

    MP3_128 = "MP3_128"
    MP3_320 = "MP3_320"
    FLAC = "FLAC"


@dataclass
class IsDataclassProtocol(Protocol):
    """A protocol to type check dataclass mixins."""


class ToListMixin(IsDataclassProtocol):
    """A dataclass mixin that converts all fields values to a list."""

    def _dataclass_to_list(self, target=None) -> List[str | List]:
        """Convert all field values to a list."""
        if target is None:
            target = asdict(self)
        return [
            v if not isinstance(v, dict) else self._dataclass_to_list(v)
            for v in target.values()
        ]

    def to_list(self, target: List[str | List] | None = None) -> List[str]:
        """Convert nested dataclasses to values list."""
        if target is None:
            target = self._dataclass_to_list()
        out = []
        for i in target:
            if isinstance(i, list):
                out += self.to_list(i)
            else:
                out.append(i)
        return out


@dataclass
class ArtistShort(ToListMixin):
    """A small model to represent an artist."""

    id: str
    name: str


@dataclass
class AlbumShort(ToListMixin):
    """A small model to represent an artist."""

    id: str
    name: str
    artist: ArtistShort


@dataclass
class TrackShort(ToListMixin):
    """A small model to represent an artist."""

    id: str
    title: str
    album: AlbumShort


Collection = List[ArtistShort] | list[AlbumShort] | List[TrackShort]


class DeezerClient(deezer.Deezer):
    """A wrapper for the Deezer API client."""

    def __init__(
        self,
        arl: str | None = None,
        quality: StreamQuality | None = None,
        fast: bool = False,
    ) -> None:
        """Instantiate the Deezer API client.

        Fast login is useful to quicky access some API endpoints such as "search" but
        won't work if you need to stream tracks.
        """
        super().__init__()

        self.arl = arl or settings.arl
        self.quality = quality or settings.quality
        if fast:
            self._fast_login()
        else:
            self._login()

    def _login(self):
        """Login to deezer API."""
        logger.debug("Login in to deezer using defined ARL…")
        self.login_via_arl(self.arl)

    def _fast_login(self):
        """Fasting login using ARL cookie."""
        cookie_obj = requests.cookies.create_cookie(
            domain=".deezer.com",
            name="arl",
            value=self.arl,
            path="/",
            rest={"HttpOnly": True},
        )
        self.session.cookies.set_cookie(cookie_obj)
        self.logged_in = True

    @staticmethod
    def _to_tracks(data) -> Generator[TrackShort, None, None]:
        """API results to TrackShort."""
        for track in data:
            yield TrackShort(
                id=str(track.get("id")),
                title=track.get("title"),
                album=AlbumShort(
                    id=str(track.get("album").get("id")),
                    name=track.get("album").get("title"),
                    artist=ArtistShort(
                        id=str(track.get("artist").get("id")),
                        name=track.get("artist").get("name"),
                    ),
                ),
            )

    def artist(
        self, artist_id: str, radio: bool = False, top: bool = True, limit: int = 10
    ) -> List[TrackShort]:
        """Get artist tracks."""
        if radio:
            response = self.api.get_artist_radio(artist_id, limit=limit)
        elif top:
            response = self.api.get_artist_top(artist_id, limit=limit)
        else:
            raise ValueError("Either radio or top should be True to get artist tracks")

        return list(self._to_tracks(response["data"]))

    def search(
        self,
        artist: str = "",
        album: str = "",
        track: str = "",
        strict: bool = False,
    ) -> Collection:
        """Mixed custom search."""
        results: Collection = []

        if len(list(filter(None, (artist, album, track)))) > 1:
            response = self.api.advanced_search(
                artist=artist, album=album, track=track, strict=strict
            )
            results = list(self._to_tracks(response["data"]))
        elif artist:
            response = self.api.search_artist(artist)
            results = [
                ArtistShort(
                    id=str(a.get("id")),
                    name=a.get("name"),
                )
                for a in response["data"]
            ]
        elif album:
            response = self.api.search_album(album)
            results = [
                AlbumShort(
                    id=str(a.get("id")),
                    name=a.get("title"),
                    artist=ArtistShort(
                        id=str(a.get("artist").get("id")),
                        name=str(a.get("artist").get("name")),
                    ),
                )
                for a in response["data"]
            ]
        elif track:
            response = self.api.search_track(track)
            results = list(self._to_tracks(response["data"]))

        return results


class TrackStatus(IntEnum):
    """Track statuses."""

    IDLE = 1
    FETCHING = 2
    PLAYABLE = 3
    FETCHED = 4


class Track:
    """A Deezer track."""

    def __init__(
        self,
        client: DeezerClient,
        track_id: str,
        quality: StreamQuality | None = None,
        buffer: float = 0.5,  # 500ms
    ) -> None:
        """Instantiate a new track."""
        self.deezer = client
        self.track_id = track_id
        self.session = requests.Session()
        self.quality = quality or self.deezer.quality
        self.track_info: dict = self._get_track_info()
        self.url: str = self._get_url()
        self.key: bytes = self._generate_blowfish_key()
        self.status: TrackStatus = TrackStatus.IDLE
        # Content and related memory view will be allocated later (right before fetching
        # the track to decrease memory footprint while adding tracks to queue).
        self.content: bytearray = bytearray()
        self._content_mv: memoryview = memoryview(self.content)
        self.fetched: int = 0
        self.streamed: int = 0
        self.paused: bool = False
        self.bitrate = self.filesize / self.duration
        self.buffer_size: int = int(self.bitrate * buffer)

    def _get_track_info(self) -> dict:
        """Get track info."""
        track_info = self.deezer.gw.get_track(self.track_id)
        logger.debug("Track info: %s", track_info)
        return track_info

    def _get_url(self) -> str:
        """Get URL of the track to stream."""
        logger.debug(f"Getting track url with quality {self.quality}…")
        url = self.deezer.get_track_url(self.token, self.quality.value)
        logger.debug(f"Track url: {url}")
        return url

    def _allocate_content(self) -> None:
        """Allocate memory where we will read/write track bytes."""
        self.content = bytearray(self.filesize)
        self._content_mv = memoryview(self.content)

    def _generate_blowfish_key(self) -> bytes:
        """Generate the blowfish key for Deezer downloads.

        Taken from: https://github.com/nathom/streamrip/
        """
        md5_hash = hashlib.md5(self.track_id.encode()).hexdigest()  # noqa: S324
        # good luck :)
        return "".join(
            chr(functools.reduce(lambda x, y: x ^ y, map(ord, t)))
            for t in zip(
                md5_hash[:16],
                md5_hash[16:],
                settings.DEEZER_BLOWFISH_SECRET,
                strict=False,
            )
        ).encode()

    def _decrypt(self, chunk):
        """Decrypt blowfish encrypted chunk."""
        return Blowfish.new(  # noqa: S304
            self.key,
            Blowfish.MODE_CBC,
            b"\x00\x01\x02\x03\x04\x05\x06\x07",
        ).decrypt(chunk)

    @property
    def token(self) -> str:
        """Get track token."""
        return self.track_info["TRACK_TOKEN"]

    @property
    def duration(self) -> int:
        """Get track duration (in seconds)."""
        return int(self.track_info["DURATION"])

    @property
    def filesize(self) -> int:
        """Get file size (in bits)."""
        return int(self.track_info[f"FILESIZE_{self.quality}"])

    @property
    def artist(self) -> str:
        """Get track artist."""
        return self.track_info["ART_NAME"]

    @property
    def title(self) -> str:
        """Get track title."""
        return self.track_info["SNG_TITLE"]

    @property
    def album(self) -> str:
        """Get track album."""
        return self.track_info["ALB_TITLE"]

    @property
    def full_title(self) -> str:
        """Get track full title (artist/title/album)."""
        return f"{self.artist} - {self.title} [{self.album}]"

    def fetch(self):
        """Fetch track in-memory.

        buffer_size (int): the buffer size (defaults to 5 seconds for a 128kbs file)
        """
        logger.debug(f"Start fetching track with {self.buffer_size=}")
        chunk_sep = 2048
        chunk_size = 3 * chunk_sep
        self.fetched = 0
        self.status = TrackStatus.IDLE
        self._allocate_content()

        with self.session.get(self.url, stream=True) as r:
            r.raise_for_status()
            filesize = int(r.headers.get("Content-Length", 0))
            logger.debug(f"Track size: {filesize} ({self.filesize})")
            self.status = TrackStatus.FETCHING

            for chunk in r.iter_content(chunk_size):
                if len(chunk) > chunk_sep:
                    dchunk = self._decrypt(chunk[:chunk_sep]) + chunk[chunk_sep:]
                else:
                    dchunk = chunk
                self._content_mv[self.fetched : self.fetched + chunk_size] = dchunk
                self.fetched += chunk_size

                if (
                    self.fetched >= self.buffer_size
                    and self.status < TrackStatus.PLAYABLE
                ):
                    logger.debug("Buffering ok")
                    self.status = TrackStatus.PLAYABLE

        # We are done here
        self.status = TrackStatus.FETCHED
        logger.debug("Track fetched")

    def cast(self, socket, chunk_size: int = 1024):
        """Cast the track via UDP using given socket."""
        multicast_group = tuple(settings.MULTICAST_GROUP)
        logger.debug(
            (
                f"Casting from position {self.streamed} with {chunk_size=} "
                f"using socket {socket} "
                f"({multicast_group=})"
            )
        )

        if self.status < TrackStatus.FETCHING:
            logger.debug("Will start fetching content in a new thead…")
            thread = Thread(target=self.fetch)
            thread.start()

        # Wait for the track to be playable
        while self.status < TrackStatus.PLAYABLE:
            sleep(0.01)

        # Sleep time while playing
        wait = 1.0 / (self.bitrate / chunk_size)
        logger.debug(f"Wait time: {wait}s ({self.bitrate=})")

        slow_connection: bool = False
        for start in range(self.streamed, self.filesize, chunk_size):

            # Track has been paused, wait for resume
            while self.paused:
                sleep(0.001)

            # We have buffering issues
            while (self.fetched - start) < self.buffer_size and start < (
                self.filesize - self.buffer_size
            ):
                if not slow_connection:
                    logger.warning(
                        "Slow connection, filling the buffer "
                        f"{self.fetched - self.streamed} < {self.buffer_size}"
                    )
                    logger.debug(
                        f"{start=} | {self.filesize=} | {self.fetched=} | "
                        f"{self.streamed=} | {chunk_size=}"
                    )
                slow_connection = True
                sleep(0.05)
            slow_connection = False

            socket.sendto(self._content_mv[start : start + chunk_size], multicast_group)
            self.streamed += chunk_size
            sleep(wait)



"""Onzr: command line interface."""

import logging
from random import shuffle
from threading import Thread
from typing import List

import click
import typer
from pynput import keyboard
from rich.console import Console
from rich.logging import RichHandler
from rich.table import Table

from .core import Onzr
from .deezer import AlbumShort, ArtistShort, Collection, StreamQuality, TrackShort

FORMAT = "%(message)s"
logging_console = Console(stderr=True)
logging.basicConfig(
    level=logging.DEBUG,
    format=FORMAT,
    datefmt="[%X]",
    handlers=[RichHandler(console=logging_console)],
)

cli = typer.Typer(name="onzr", no_args_is_help=True, pretty_exceptions_short=True)
console = Console()
logger = logging.getLogger(__name__)


def start(fast: bool = False, quiet: bool = False) -> Onzr:
    """Start onzr."""
    if not quiet:
        console.print("🚀 login in to Deezer…", style="cyan")
    return Onzr(fast=fast)


def print_collection_ids(collection: Collection):
    """Print a collection as ids."""
    for item in collection:
        console.print(item.id)


def print_collection_table(collection: Collection, title="Collection"):
    """Print a collection as a table."""
    table = Table(title=title)

    sample = collection[0]
    show_artist = (
        True
        if isinstance(sample, TrackShort)
        or isinstance(sample, AlbumShort)
        or isinstance(sample, ArtistShort)
        else False
    )
    show_album = (
        True
        if isinstance(sample, TrackShort) or isinstance(sample, AlbumShort)
        else False
    )
    show_track = True if isinstance(sample, TrackShort) else False

    if show_track:
        table.add_column("ID", justify="right")
        table.add_column("Track", style="#9B6BDF")
    if show_album:
        table.add_column("ID", justify="right")
        table.add_column("Album", style="#E356A7")
    if show_artist:
        table.add_column("ID", justify="right")
        table.add_column("Artist", style="#75D7EC")

    for item in collection:
        table.add_row(*item.to_list())

    console.print(table)


@cli.command()
def search(  # noqa: PLR0913
    artist: str = "",
    album: str = "",
    track: str = "",
    strict: bool = False,
    quiet: bool = False,
    ids: bool = False,
):
    """Search track, artist and/or album."""
    if ids:
        quiet = True
    onzr = start(fast=True, quiet=quiet)
    if not quiet:
        console.print("🔍 start searching…")

    results = onzr.deezer.search(artist, album, track, strict)

    if not results:
        console.print("No match found.")
        typer.Exit(code=1)

    if ids:
        print_collection_ids(results)
        return

    print_collection_table(results, title="Search results")


@cli.command()
def artist(  # noqa: PLR0913
    artist_id: str,
    top: bool = True,
    radio: bool = False,
    limit: int = 10,
    quiet: bool = True,
    ids: bool = False,
):
    """Get artist popular track ids."""
    if not top and not radio:
        console.print("You should choose either top titles or artist radio.")
        raise typer.Exit(code=2)

    if ids:
        quiet = True

    if artist_id == "-":
        logger.debug("Reading artist id from stdin…")
        artist_id = click.get_text_stream("stdin").read().strip()
        logger.debug(f"{artist_id=}")

    onzr = start(fast=True, quiet=quiet)
    tracks = onzr.deezer.artist(artist_id, radio=radio, top=top, limit=limit)

    if ids:
        print_collection_ids(tracks)
        return

    print_collection_table(tracks, title="Artist tracks")


@cli.command()
def mix(
    artist: list[str],
    deep: bool = False,
    limit: int = 10,
    quiet: bool = True,
    ids: bool = False,
):
    """Create a playlist from multiple artists."""
    if ids:
        quiet = True

    onzr = start(fast=True, quiet=quiet)
    tracks = []

    if not quiet:
        console.print("🍪 cooking the mix…")

    for artist_ in artist:
        result = onzr.deezer.search(artist_, strict=True)
        # We expect the search engine to be relevant 🤞
        artist_id = result[0].id
        tracks += onzr.deezer.artist(artist_id, radio=deep, top=True, limit=limit)
    shuffle(tracks)

    if ids:
        print_collection_ids(tracks)
        return

    print_collection_table(tracks, title="Onzr Mix tracks")


@cli.command()
def play(
    track_ids: List[str],
    quality: StreamQuality = StreamQuality.MP3_128,
    shuffle: bool = False,
):
    """Play one (or more) tracks."""
    onzr = start()
    console.print("🚀 starting the player…")
    if track_ids == ["-"]:
        logger.debug("Reading track ids from stdin…")
        track_ids = click.get_text_stream("stdin").read().split()
        logger.debug(f"{track_ids=}")
    onzr.add(track_ids, quality)
    if shuffle:
        onzr.shuffle()

    # Start playing in a new thread
    thread = Thread(target=onzr.play)
    thread.start()

    # Controls
    def on_press(key: keyboard.Key):
        """Player control actions."""
        match key:
            case keyboard.Key.media_play_pause:
                onzr.pause()

    with keyboard.Listener(on_press=on_press) as listener:  # type: ignore[arg-type]
        listener.join()

    typer.Exit()



"""Onzr: core module."""

import logging
import random
import socket
import struct
from enum import IntEnum
from socket import SocketType
from typing import List

from .deezer import DeezerClient, StreamQuality, Track
from .player import Player

logger = logging.getLogger(__name__)


class Queue:
    """Onzr playing queue."""

    def __init__(self) -> None:
        """Instantiate the tracks queue."""
        self.current: Track | None = None
        self.tracks: List[Track] = []

    def __len__(self):
        """Get queue length."""
        return len(self.tracks)

    @property
    def is_empty(self):
        """Check if tracks are queued."""
        return len(self) == 0 and self.current is None

    def add(self, track: Track | None = None, tracks: List[Track] | None = None):
        """Add one or more tracks to queue."""
        if track is None and tracks is None:
            raise TypeError("Argument missing, you should either add a track or tracks")
        self.tracks.extend(tracks or [track])  # type: ignore[list-item]

    def shuffle(self):
        """Shuffle current track list."""
        random.shuffle(self.tracks)

    def next(self):
        """Set next track as the current."""
        del self.current
        if not len(self):
            self.current = None
            return
        self.current = self.tracks.pop(0)


class OnzrStatus(IntEnum):
    """Onzr player status."""

    IDLE = 1
    PLAYING = 2
    PAUSED = 3
    STOPPED = 4


class Onzr:
    """Onzr core class."""

    def __init__(self, fast: bool = False) -> None:
        """Initialize all the things.

        fast (bool): activate Deezer fast login
        """
        logger.debug("Instantiating Onzr…")

        self.deezer: DeezerClient = DeezerClient(fast=fast)
        self.socket: SocketType = self.configure_socket()
        self.player: Player = Player(self.socket)
        self._queue: Queue = Queue()
        self.status: OnzrStatus = OnzrStatus.IDLE

    def configure_socket(self):
        """Open and configure the casting socket."""
        logger.debug("Setting socket…")
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)

        # Configure TTL
        ttl = struct.pack("b", 1)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, ttl)
        logger.debug(f"Socket: {sock}")
        return sock

    def add(self, track_ids: List[str], quality: StreamQuality = StreamQuality.MP3_320):
        """Little helper to queue tracks."""
        logger.info("➕ Adding new tracks to queue…")
        tracks = [Track(self.deezer, track_id, quality) for track_id in track_ids]
        for track in tracks:
            logger.info(f"{track.full_title}")
        self._queue.add(tracks=tracks)
        logger.info(f"✅ {len(tracks)} tracks queued")

    def shuffle(self):
        """Shuffle tracks queue."""
        logger.info("🔀 Shuffling tracks queue…")
        self._queue.shuffle()
        for track in self._queue.tracks:
            logger.info(f"{track.full_title}")

    def play(self):
        """Little helper to start playing the queue."""
        # No active track, try to queue the next one in line
        if self._queue.current is None:
            self._queue.next()

        # There is no track left in queue
        if self._queue.is_empty:
            return

        # Play current track
        self.status = OnzrStatus.PLAYING
        self._queue.current.paused = False
        self.player.play(self._queue.current)
        self._queue.next()

        return self.play()

    def pause(self):
        """Pause current track playing."""
        logger.info("⏯ Toggling pause…")
        logger.debug(f"Onzr Status? {self.status.name}")
        if self._queue.current is None:
            return

        if self.status == OnzrStatus.PLAYING:
            self.status = OnzrStatus.PAUSED
            self._queue.current.paused = True
        elif self.status == OnzrStatus.PAUSED:
            self.status = OnzrStatus.PLAYING
            self._queue.current.paused = False



"""Onzr: player."""

import logging
from socket import SocketType

import vlc
from vlc import MediaPlayer

from .config import settings
from .deezer import Track

logger = logging.getLogger(__name__)


class Player:
    """Onzr player."""

    def __init__(self, socket: SocketType, local: bool = True) -> None:
        """Instantiate the player."""
        logger.debug(f"Instantiating Player using socket: {socket} ({local=})")
        self.socket: SocketType = socket
        self.track: Track | None = None
        self.chunk_size: int = 1024 * 16
        self.vlc: MediaPlayer | None = None
        if local:
            self._start_udp_client(settings.MULTICAST_URL)

    def _start_udp_client(self, url: str):
        """Start local UDP client.

        Start a local UDP client that will listen to the UDP socket and play sound.
        """
        logger.debug(f"Starting local UDP client with: {url=}")
        self.vlc = vlc.MediaPlayer(url)
        self.vlc.play()

    def play(self, track: Track):
        """Play a track."""
        logger.info(f"▶️ {track.full_title}")
        self.track = track
        self.track.cast(self.socket, self.chunk_size)
